#include <sstream>

#include <cstdlib>
#include <cstring>

#include <gtkmm.h>

#include "gui-main.h"
#include "parse.h"

#include <iostream>

#include <readline/readline.h>
#include <readline/history.h>

#include <gtkmm.h>

static int available_char = 0;

static inline int
ctrl (int c)
{
  return c & 0x1f;
}

static char *
strsave (const char *s)
{
  if (! s)
    return 0;

  int len = strlen (s);
  char *tmp = new char [len+1];
  tmp = strcpy (tmp, s);
  return tmp;
}

class command_window : public Gtk::Window
{
public:

  command_window (void)
    : m_VBox (Gtk::ORIENTATION_VERTICAL), scrolled_window (),
      m_TextView (), buffer (), beg_mark (), prompt_mark ()
  {
    set_title ("Gtk::TextView example");
    set_border_width (5);
    set_default_size (600, 400);

    add (m_VBox);

    // Add the TreeView, inside a ScrolledWindow, with the button
    // underneath:

    scrolled_window.add (m_TextView);

    // Only show the scrollbars when they are necessary:

    scrolled_window.set_policy (Gtk::POLICY_AUTOMATIC, Gtk::POLICY_ALWAYS);

    m_VBox.pack_start (scrolled_window);

    buffer = Gtk::TextBuffer::create ();
    buffer->set_text
      ("Gtk Example Calculator.\n"
       "Available operations: + - * / ^ ()\n"
       "Semicolon terminates statement.\n"
       "Up Arrow key moves to previous line in the command history.\n"
       "Down Arrow key moves to next line in the comand history.\n\n");
    m_TextView.set_buffer (buffer);

    beg_mark = set_mark ();

    show_all_children ();
  }

  virtual ~command_window (void) { }

  // Accept an input line, parse and possibly execute it.

  void accept (const std::string& line)
  {
    insert_at_end ("\n");

    if (! line.empty ())
      {
        add_history (line.c_str ());
        using_history ();

        int status = parse_and_execute (line);

        if (have_result)
          {
            insert_result ();

            have_result = false;
          }
      }
  }

  // Redisplay current command line.

  void redisplay (void)
  {
    erase_line ();

    std::string line = rl_line_buffer ? rl_line_buffer : "";
    std::string prompt = (rl_prompt && beg_of_stmt) ? rl_prompt : "";

    insert_line (prompt, line);

    Gtk::TextBuffer::iterator iter = buffer->get_iter_at_mark (prompt_mark);

    for (int i = 0; i < rl_point; i++)
      iter++;

    scroll_to_bottom ();

    buffer->place_cursor (iter);
  }

  void insert_at_end (const std::string& text)
  {
    Gtk::TextBuffer::iterator iter = buffer->insert (buffer->end (), text);

    scroll_to_bottom ();

    beg_mark = set_mark ();
  }

protected:

  bool on_key_press_event (GdkEventKey* event)
  {
    bool retval = false;

    if (! event)
      return retval;

    if (event->type == GDK_KEY_PRESS)
      {
        int key = -1;

        if (event->keyval > 0x19 && event->keyval < 0x7f)
          {
            key = event->keyval;

            if ((event->state & GDK_CONTROL_MASK)
                && key > 0x3f && key < 0x7b)
              key &= 0x1f;
          }
        else
          {
            switch (event->keyval)
              {
              case GDK_KEY_Return:
                key = 0x0A;
                break;

              case GDK_KEY_BackSpace:
                key = 0x08;
                break;

              case GDK_KEY_Tab:
                key = 0x09;
                break;

              case GDK_KEY_Escape:
                key = 0x1b;
                break;

              case GDK_KEY_Up:
              case GDK_KEY_Down:
              case GDK_KEY_Right:
              case GDK_KEY_Left:
                key = do_arrow_key (event->keyval);
                break;

              default:
                break;
              }
          }

        if (key >= 0)
          {
            available_char = key;
            rl_callback_read_char ();
            retval = true;
          }
        else
          {
            // Unhandled key.
            retval = true;
          }
      }

  done:

    if (! retval)
      Gtk::Widget::on_key_press_event (event);

    return retval;
  }

  int do_arrow_key (int arrow_key)
  {
    int retval = 0;

    available_char = 0x1b;
    rl_callback_read_char ();

    available_char = '[';
    rl_callback_read_char ();

    switch (arrow_key)
      {
      case GDK_KEY_Up:
        retval = 'A';
        break;

      case GDK_KEY_Down:
        retval = 'B';
        break;

      case GDK_KEY_Right:
        retval = 'C';
        break;

      case GDK_KEY_Left:
        retval = 'D';
        break;
      }

    return retval;
  }

  // Retrieve a command from the history list and insert it on the
  // current command.

  void history (bool up)
  {
    HIST_ENTRY *entry = up ? previous_history () : next_history ();

    if (entry)
      {
        erase_line ();

        std::string prompt = rl_prompt ? rl_prompt : "";

        insert_line (prompt, entry->line);
      }
    else if (! up)
      erase_line ();

    scroll_to_bottom ();
  }

  void erase_line (void)
  {
    Glib::RefPtr<Gtk::TextBuffer::Mark> mark = buffer->get_insert ();

    Gtk::TextBuffer::iterator beg_iter
      = buffer->get_iter_at_mark (beg_mark);

    buffer->erase (beg_iter, buffer->end ());
  }

  void insert_line (const std::string& prompt, const std::string& line)
  {
    beg_mark = set_mark ();

    buffer->insert_at_cursor (prompt);

    prompt_mark = set_mark ();

    buffer->insert_at_cursor (line);
  }

  Glib::RefPtr<Gtk::TextBuffer::Mark> set_mark (void)
  {
    // We can't just use this because the mark associated with the
    // cursor moves with the cursor.  But is there a simpler way to do
    // this job?

    Glib::RefPtr<Gtk::TextBuffer::Mark> cursor = buffer->get_insert ();

    Gtk::TextBuffer::iterator iter = buffer->get_iter_at_mark (cursor);

    return buffer->create_mark (iter);
  }

  void insert_result (void)
  {
    std::ostringstream buf;
    buf << "ans = " << result << "\n";

    buffer->insert_at_cursor (buf.str ());

    beg_mark = set_mark ();

    scroll_to_bottom ();
  }

  void scroll_to_bottom (void)
  {
    Glib::RefPtr<Gtk::Adjustment> adj = scrolled_window.get_vadjustment ();

    adj->set_value (adj->get_upper ());

    while (Gtk::Main::events_pending ())
      Gtk::Main::iteration ();
  }

  // Child widgets:

  Gtk::VBox m_VBox;

  Gtk::ScrolledWindow scrolled_window;
  Gtk::TextView m_TextView;

  Glib::RefPtr<Gtk::TextBuffer> buffer;

  Glib::RefPtr<Gtk::TextBuffer::Mark> beg_mark;
  Glib::RefPtr<Gtk::TextBuffer::Mark> prompt_mark;
};

static command_window *calc_interaction_window = 0;

static int
gui_getc (FILE *)
{
  int tmp = available_char;
  available_char = 0;
  return tmp;
}

static void
gui_redisplay (void)
{
  if (calc_interaction_window)
    calc_interaction_window->redisplay ();
}

static void
gui_prep_term (int)
{
}

static void
gui_deprep_term (void)
{
}

static void
gui_accept_line (char *line)
{
  if (calc_interaction_window)
    calc_interaction_window->accept (line);
}

static void
gui_display_completion_matches (char **matches, int num_matches,
                                int /* max_length */)
{
  if (calc_interaction_window)
    {
      std::ostringstream buf;

      if (num_matches > 1)
        buf << "\n";

      for (int i = 1; i < num_matches; i++)
        buf << matches[i] << "\n";

      calc_interaction_window->insert_at_end (buf.str ());

      calc_interaction_window->redisplay ();
    }
}

static void
readline_init (void)
{
  rl_initialize ();

  rl_getc_function = gui_getc;
  rl_redisplay_function = gui_redisplay;
  rl_prep_term_function = gui_prep_term;
  rl_deprep_term_function = gui_deprep_term;
  rl_completion_display_matches_hook = gui_display_completion_matches;

  rl_callback_handler_install (">> ", gui_accept_line);
}

static void
readline_fini (void)
{
  rl_callback_handler_remove ();
}

int
gui_main (int argc, char *argv[])
{
  parser_init ();

  Gtk::Main kit (argc, argv);

  calc_interaction_window = new command_window ();

  readline_init ();

  // Shows the window and returns when it is closed.
  Gtk::Main::run (*calc_interaction_window);

  readline_fini ();

  parser_fini ();

  return 0;
}

// -- Variable: rl_getc_func_t * rl_getc_function
//     If non-zero, Readline will call indirectly through this pointer to
//     get a character from the input stream.  By default, it is set to
//     `rl_getc', the default Readline character input function (*note
//     Character Input::).

// -- Variable: rl_voidfunc_t * rl_redisplay_function
//     If non-zero, Readline will call indirectly through this pointer to
//     update the display with the current contents of the editing buffer.
//     By default, it is set to `rl_redisplay', the default Readline
//     redisplay function (*note Redisplay::).

// -- Variable: rl_vintfunc_t * rl_prep_term_function
//     If non-zero, Readline will call indirectly through this pointer to
//     initialize the terminal.  The function takes a single argument, an
//     `int' flag that says whether or not to use eight-bit characters.
//     By default, this is set to `rl_prep_terminal' (*note Terminal
//     Management::).

// -- Variable: rl_voidfunc_t * rl_deprep_term_function
//     If non-zero, Readline will call indirectly through this pointer to
//     reset the terminal.  This function should undo the effects of
//     `rl_prep_term_function'.  By default, this is set to
//     `rl_deprep_terminal' (*note Terminal Management::).

// -- Function: void rl_callback_handler_install (const char *prompt,
//          rl_vcpfunc_t *lhandler)
//     Set up the terminal for readline I/O and display the initial
//     expanded value of PROMPT.  Save the value of LHANDLER to use as a
//     function to call when a complete line of input has been entered.
//     The function takes the text of the line as an argument.

// -- Function: void rl_callback_read_char (void)
//     Whenever an application determines that keyboard input is
//     available, it should call `rl_callback_read_char()', which will
//     read the next character from the current input source.  If that
//     character completes the line, `rl_callback_read_char' will invoke
//     the LHANDLER function saved by `rl_callback_handler_install' to
//     process the line.  Before calling the LHANDLER function, the
//     terminal settings are reset to the values they had before calling
//     `rl_callback_handler_install'.  If the LHANDLER function returns,
//     the terminal settings are modified for Readline's use again.
//     `EOF' is  indicated by calling LHANDLER with a `NULL' line.

// -- Function: void rl_callback_handler_remove (void)
//     Restore the terminal to its initial state and remove the line
//     handler.  This may be called from within a callback as well as
//     independently.  If the LHANDLER installed by
//     `rl_callback_handler_install' does not exit the program, either
//     this function or the function referred to by the value of
//     `rl_deprep_term_function' should be called before the program
//     exits to reset the terminal settings.

// -- Variable: rl_compdisp_func_t * rl_completion_display_matches_hook
//     If non-zero, then this is the address of a function to call when
//     completing a word would normally display the list of possible
//     matches.  This function is called in lieu of Readline displaying
//     the list.  It takes three arguments: (`char **'MATCHES, `int'
//     NUM_MATCHES, `int' MAX_LENGTH) where MATCHES is the array of
//     matching strings, NUM_MATCHES is the number of strings in that
//     array, and MAX_LENGTH is the length of the longest string in that
//     array.  Readline provides a convenience function,
//     `rl_display_match_list', that takes care of doing the display to
//     Readline's output stream.  That function may be called from this
//     hook.
