// Infix notation calculator.

%{

#define YYSTYPE double

#include <iostream>

#include <cctype>
#include <cmath>
#include <cstdio>

#include "parse.h"

FILE *outfile = 0;
size_t bufptr = 0;
size_t chunk_size = 0;
const char *buf;
bool beg_of_stmt = true;
bool have_result = false;
double result;

static int yylex (void);
static void yyerror (char const *);

%}

%define api.push-pull push

// Bison declarations.
%token NUM
%left '-' '+'
%left '*' '/'
%left NEG     // negation--unary minus
%right '^'    // exponentiation

%%

input   : // empty
          { have_result = false; }
        | input line
        ;

line    :     ';'
          { have_result = false; }
        | exp ';'
          {
            have_result = true;
            result = $1;
            beg_of_stmt = true;
          }
;

exp     : NUM
          {
            $$ = $1;
            beg_of_stmt = false;
          }
        | exp '+' exp
          {
            $$ = $1 + $3;
            beg_of_stmt = false;
          }
        | exp '-' exp
          {
            $$ = $1 - $3;
            beg_of_stmt = false;
          }
        | exp '*' exp
          {
            $$ = $1 * $3;
            beg_of_stmt = false;
          }
        | exp '/' exp
          {
            $$ = $1 / $3;
            beg_of_stmt = false;
          }
        | '-' exp  %prec NEG
          {
            $$ = -$2;
            beg_of_stmt = false;
          }
        | exp '^' exp
          {
            $$ = std::pow ($1, $3);
            beg_of_stmt = false;
          }
        | '(' exp ')'
          {
            $$ = $2;
            beg_of_stmt = false;
          }
        ;

%%

// The lexical analyzer returns a double floating point number on the
// stack and the token NUM, or the numeric code of the character read
// if not a number.  It skips all blanks and tabs, and returns 0 for
// end-of-input.

int
static yylex (void)
{
  int c;

  if (bufptr >= chunk_size)
    return -1;

  // Skip white space.
  while ((c = buf[bufptr++]) == ' ' || c == '\t' || c == '\n')
    ;

  // Process numbers.
  if (c == '.' || isdigit (c))
    {
      int chars_read = 0;
      bufptr--;
      sscanf (&buf[bufptr], "%lf%n", &yylval, &chars_read);
      bufptr += chars_read;
      return NUM;
    }

  // Return a single char.
  return c;
}

static void
yyerror (char const *msg)
{
  std::cerr << "parse error: " << msg << std::endl;
}

static yypstate *ps = 0;

void
parser_fini (void)
{
  if (ps)
    yypstate_delete (ps);

  ps = 0;
}

void
parser_init (void)
{
  parser_fini ();

  ps = yypstate_new ();
}

int
parse_and_execute (const std::string& line)
{
  extern int yychar;

  bufptr = 0;
  chunk_size = line.length ();
  buf = line.c_str ();

  int status;

  do
    {
      yychar = yylex ();

      if (yychar < 0)
        return 0;

      status = yypush_parse (ps);
    }
  while (status = YYPUSH_MORE);

  return -1;
}
